package ru.dbotthepony.mc.otm.core

import it.unimi.dsi.fastutil.ints.IntList
import net.minecraft.Util
import net.minecraft.util.RandomSource
import ru.dbotthepony.mc.otm.core.math.Decimal
import java.math.BigInteger
import java.util.*
import java.util.random.RandomGenerator
import kotlin.experimental.and
import kotlin.math.ln
import kotlin.math.sqrt

data class DoublePair(val first: Double, val second: Double)

fun RandomSource.nextUUID(): UUID {
	return UUID(nextLong(), nextLong())
}

// normal distribution via Marsaglia polar method
fun RandomGenerator.nextNormalDoubles(stddev: Double, mean: Double): DoublePair {
	var rand1: Double
	var rand2: Double
	var distSqr: Double

	do {
		rand1 = 2.0 * nextDouble() - 1.0
		rand2 = 2.0 * nextDouble() - 1.0
		distSqr = rand1 * rand1 + rand2 * rand2
	} while (distSqr >= 1.0 || distSqr == 0.0)

	val mapping = sqrt(-2.0 * ln(distSqr) / distSqr)

	return DoublePair(
		rand1 * mapping * stddev + mean,
		rand2 * mapping * stddev + mean
	)
}

// All Mojang's random sources use MarsagliaPolarGaussian for generating normal distributed doubles
fun RandomSource.nextNormalDoubles(stddev: Double, mean: Double): DoublePair {
	return DoublePair(
		nextGaussian() * stddev + mean,
		nextGaussian() * stddev + mean
	)
}

fun RandomSource.nextNormalDouble(stddev: Double, mean: Double): Double {
	return nextGaussian() * stddev + mean
}

fun RandomSource.nextFloat(min: Float, max: Float): Float {
	if (this is RandomGenerator)
		return nextFloat(min, max)

	require(max >= min) { "Min is bigger than max: $min vs $max" }
	if (min == max) return min
	return min + nextFloat() * (max - min)
}

fun RandomSource.nextDouble(min: Double, max: Double): Double {
	if (this is RandomGenerator)
		return nextDouble(min, max)

	require(max >= min) { "Min is bigger than max: $min vs $max" }
	if (min == max) return min
	return min + nextDouble() * (max - min)
}

fun IntArray.shuffle(random: RandomSource): IntArray {
	for (i in lastIndex downTo 1) {
		val j = random.nextInt(i + 1)
		val copy = this[i]
		this[i] = this[j]
		this[j] = copy
	}

	return this
}

fun <L : IntList> L.shuffle(random: RandomSource): L {
	for (i in lastIndex downTo 1) {
		val j = random.nextInt(i + 1)
		set(j, set(i, getInt(j)))
	}

	return this
}

fun <T, L : MutableList<T>> L.shuffle(random: RandomSource): L {
	Util.shuffle(this, random)
	return this
}

fun <T> List<T>.random(random: RandomGenerator): T {
	if (isEmpty())
		throw NoSuchElementException("This list is empty")
	return get(random.nextInt(size))
}

fun <T> List<T>.random(random: RandomSource): T {
	if (isEmpty())
		throw NoSuchElementException("This list is empty")
	return get(random.nextInt(size))
}

class RandomByteSource(private val source: RandomSource) {
	private val bytes = ByteArray(8)
	private var i = 7

	fun next(): Byte {
		if (++i == 8) {
			i = 0

			var generate = source.nextLong()

			for (i in 0 .. 7) {
				bytes[i] = generate.toByte()
				generate = generate ushr 8
			}
		}

		return bytes[i]
	}

	fun next(bound: Int): Byte {
		require(bound > 0) { "Bound must be positive" }
		val m = bound - 1
		var r = next().toInt().and(0xFF)

		if (bound and m == 0) {
			r = r and m
		} else {
			var u = r ushr 1

			while (true) {
				r = u % bound

				if (u + m - r < 0) {
					u = next().toInt().and(0xFF).ushr(1)
				} else {
					break
				}
			}
		}

		return r.toByte()
	}

	fun next(bytes: ByteArray) {
		for (i in bytes.indices) {
			bytes[i] = next()
		}
	}
}

/**
 * Uniformely distributed [Decimal] value on [0,[bound]) range
 *
 * If [round] is `true`, will only return integers
 */
fun RandomSource.nextDecimal(bound: Decimal, round: Boolean = false): Decimal {
	if (round)
		require(bound > Decimal.ZERO) { "Bound must be positive, $bound given" }
	else
		require(bound >= Decimal.ONE) { "Bound must be 1 or bigger, $bound given" }

	require(bound.isFinite) { "Bound must be finite" }

	val bytes = RandomByteSource(this)

	if (round) {
		val thisBytes = bound.whole.toByteArray()
		val generateBytes = ByteArray(thisBytes.size)
		bytes.next(generateBytes)

		generateBytes[0] = generateBytes[0].and(127)

		if (generateBytes[0] >= thisBytes[0]) {
			generateBytes[0] = bytes.next(thisBytes[0].toInt().and(127))
		}

		return Decimal(BigInteger(generateBytes))
	} else {
		val thisBytes = bound.mag.toByteArray()
		val generateBytes = ByteArray(thisBytes.size)
		bytes.next(generateBytes)

		generateBytes[0] = generateBytes[0].and(127)

		if (generateBytes[0] >= thisBytes[0]) {
			generateBytes[0] = bytes.next(thisBytes[0].toInt().and(127))
		}

		return Decimal.raw(BigInteger(generateBytes))
	}
}

/**
 * Uniformely distributed [Decimal] value on [[origin],[bound]) range
 *
 * If [round] is `true`, will only return integers
 */
fun RandomSource.nextDecimal(origin: Decimal, bound: Decimal, round: Boolean = false): Decimal {
	require(origin < bound) { "Origin must be less than bound: $origin < $bound" }
	require(origin.isFinite) { "Origin must be finite" }
	require(bound.isFinite) { "Bound must be finite" }

	return origin + nextDecimal(bound - origin, round)
}

/**
 * Uniformely distributed [Decimal] value on [0,1) range
 */
fun RandomSource.nextDecimal(): Decimal {
	return nextDecimal(Decimal.ZERO, Decimal.ONE)
}

fun RandomSource.nextVariance(value: Decimal, round: Boolean = false): Decimal {
	return nextDecimal(-value / 2, value / 2, round)
}
